#include <R.h>
#include <Rmath.h>
#include <Rinternals.h>

#define DIMS(x)							\
  INTEGER(coerceVector(getAttrib((x), R_DimSymbol), INTSXP))

/* wtnum computes the numerator of the weight for permutation i 
(ith row in permutation matrix) in the likelihood function*/  
double wtnum(int *base, int irow, int *a, int *na, double *theta)
{
  int j, diff = 0;
  double res = 0.0;
  for (j = 0; j < na[1]; j++) {
	diff += abs(base[j] - a[irow + j*na[0]]);
  }
  diff *= -1;
  res = exp((*theta) * diff); 
  return res;
}

/* wtden calculates the denominator of the weight for permutation i 
(ith row in permutation matrix) in the likelihood function*/  
double wtden(int *base, int *a, int *na, double *theta)
{
  int irow;
  double sum, tmp;
  sum = 0.0;
  tmp = 0.0;
  for (irow = 0; irow < na[0]; irow++) {
	tmp = wtnum(base, irow, a, na, theta);
	sum += tmp;
  }
  return sum;
}

/* thetafun returns the theta in the weights computation*/
/* theta could be obtained in upper level - R interface and 
   used as par in log sum function */
/*
double thetafun(int *n)
{ 
  double theta = 0.0;
   theta = 1/n; */
  /* use the function passed from R/user provided  
  theta = rtheta(n);
  return theta;	
}
*/

/* logsum computes the log of the weighted sum within each 
individual in the likelihood function */
double logsum(double *P, int *a, int *nP, int *na, int *base, 
          double *theta)
{
  int irow, j;
  double tmp, sum, iwtnum, iwtden;
  sum = 0.0;
  iwtden = wtden(base, a, na, theta);
  for (irow = 0; irow < na[0]; irow++) {
    tmp = 1.0;
    for (j = 0; j < na[1]; j++) {
      tmp *= P[a[irow + j*na[0]] - 1 + j*nP[0]];
	}
	iwtnum = wtnum(base, irow, a, na, theta);
    sum += iwtnum / iwtden * tmp;
  }
  return log(sum);
} 

/* logsumomit computes the log of the weighted sum within each 
individual in the likelihood function. if loop - not so efficient*/ 
double logsumomit(double *P, int *a, int *nP, int *na, int *base, 
              double *theta, int romit, int comit)
{
  int i, j;
  double tmp, sum, iwtnum, iwtden;
  sum = 0.0;
  iwtden = wtden(base, a, na, theta);
  for (i = 0; i < na[0]; i++) {
    tmp = 1.0;
    iwtnum = wtnum(base, i, a, na, theta);
	if (a[i + comit*na[0]] - 1 == romit) {
	  for (j = 0; j < na[1]; j++) {
	    if (j == comit) {
		  tmp *= 1.0;
		} else {
		  tmp *= P[a[i + j*na[0]] - 1 + j*nP[0]];
		}
	  }
	} else {
		tmp = 0.0;
	}
    /* for (j = 0; j < na[1]; j++) {
	  if (j == comit) {
		if (a[i + j*na[0]] - 1 == romit) {
		  tmp *= 1.0;
		} else {
		  tmp *= 0.0;
		} 
	  } else {
		  tmp *= P[a[i + j*na[0]] - 1 + j*nP[0]];
	  }
	} */
    sum += iwtnum / iwtden * tmp;
  }
  return log(sum);
} 

/* dvecsum returns the sum of elements in vector x */
double dvecsum(double *x, int n)
{
  int i;
  double sum = 0.0;
  for (i = 0; i < n; i++)
    sum += x[i];
  return sum;    
}

/* irowsum returns the sum of elements in row row of x */
int irowsum(int *x, int *nx, int row)
{
  int i, sum = 0;
  for (i = 0; i < nx[1]; i++) {
    sum += x[row + i*nx[0]];
  }
  return sum;    
}

/* get element i of vector */
/*double veci (double *vector, int n)
{
  return vector[n];
}*/

/* dsubmat makes a substitution for the rows and columns */
void dsubmat(double *old, double *new, int *nold, int *nnew,
	     int *rowold, int *colold)
{
  int i, j;
  if (rowold != NULL && colold == NULL) {
    for (i = 0; i < nnew[0]; i++)
      for (j = 0; j < nnew[1]; j++)
	    new[i + j*nnew[0]] = old[rowold[i] + j*nold[0]];
  } else if (rowold == NULL && colold != NULL) {
    for (i = 0; i < nnew[0]; i++)
      for (j = 0; j < nnew[1]; j++)
	    new[i + j*nnew[0]] = old[i + colold[j]*nold[0]];
  } else if (rowold != NULL && colold != NULL) {
    for (i = 0; i < nnew[0]; i++)
      for (j = 0; j < nnew[1]; j++)
	    new[i + j*nnew[0]] = old[rowold[i] + colold[j]*nold[0]];
  } else {
    for (i = 0; i < nnew[0]; i++)
      for (j = 0; j < nnew[1]; j++)
	    new[i + j*nnew[0]] = old[i + j*nold[0]];
  }
}

/* isubmat makes a substitution for the rows and columns */
void isubmat(int *old, int *new, int *nold, int *nnew,
	     int *rowold, int *colold)
{
  int i, j;
  if (rowold != NULL && colold == NULL) {
    for (i = 0; i < nnew[0]; i++)
      for (j = 0; j < nnew[1]; j++)
	    new[i + j*nnew[0]] = old[rowold[i] + j*nold[0]];
  } else if (rowold == NULL && colold != NULL) {
    for (i = 0; i < nnew[0]; i++)
      for (j = 0; j < nnew[1]; j++)
	    new[i + j*nnew[0]] = old[i + colold[j]*nold[0]];
  } else if (rowold != NULL && colold != NULL) {
    for (i = 0; i < nnew[0]; i++)
      for (j = 0; j < nnew[1]; j++)
	    new[i + j*nnew[0]] = old[rowold[i] + colold[j]*nold[0]];
  } else {
    for (i = 0; i < nnew[0]; i++)
      for (j = 0; j < nnew[1]; j++)
	    new[i + j*nnew[0]] = old[i + j*nold[0]];
  }
}

/* loglik function computes the log-likelihood of the model */
SEXP loglik(SEXP x, SEXP AllPtrue, SEXP divsel, 
	    SEXP Y, SEXP a, SEXP MAX, SEXP base, SEXP theta)
{
  int nx, ndivsel, *nAll, *nY, *nP, *na_N;
  int *rY, *rdivsel, *ra_N, *rbase_i;
  /* int * rMAX; */
  int i, j, k, l, N, *rowAll;

  double *rx, *rAllPtrue, *rans, *rtheta_i;
  double *AllP, *P, *loglik;
  /* double *rtheta; */
 
  SEXP ans;

  PROTECT(ans = allocVector(REALSXP, 1));
  rans = REAL(ans);

  nx	 = length(x);
  ndivsel = length(divsel);
  nAll = DIMS(AllPtrue);
  nY	 = DIMS(Y);
  nP = (int *)R_alloc(2, sizeof(int));

  rx = REAL(x);
  rAllPtrue = REAL(AllPtrue);
  /* rtheta = REAL(theta); */
  rdivsel = INTEGER(divsel); 
  rY = INTEGER(Y);
  /* rMAX = INTEGER(MAX); */
  /* rbase = INTEGER(base); */

  loglik = (double *)R_alloc(nY[0], sizeof(double));
  /* div_sel = INTEGER(VECTOR_ELT(division, rsel[0]-1))[0]; */

  AllP = (double *)R_alloc(nAll[0]*nAll[1], sizeof(double));
  for (i = 0; i < nAll[0]; i++)
    for (j = 0; j < nAll[1]; j++)
      AllP[i + j*nAll[0]] = rAllPtrue[i + j*nAll[0]];

  for(j = 0; j < ndivsel; j++) {
    for (i = 0; i < nAll[0] - 1; i++) {
	  AllP[i + (rdivsel[j] - 1)*nAll[0]] = rx[i];		  
	}
    AllP[i + (rdivsel[j] - 1)*nAll[0]] = 1 - dvecsum(rx, nx);
  }

  for (i = 0; i < nY[0]; i++) {

    N = irowsum(rY, nY, i);

    nP[0] = N;
    nP[1] = N;
    P     = (double *)R_alloc(nP[0]*nP[1], sizeof(double));
    rowAll = (int *)R_alloc(nP[0], sizeof(int));
	
    l = 0;
    for (j = 0; j < nY[1]; j++) {
      for (k = 0; k < rY[i + j*nY[0]]; k++) {
	    rowAll[l++] = j;
      }
	}

    dsubmat(AllP, P, nAll, nP, rowAll, NULL);
	
    na_N = DIMS(VECTOR_ELT(a, N-1));
    ra_N = INTEGER(VECTOR_ELT(a, N-1));
	rbase_i = INTEGER(VECTOR_ELT(base, i));
	/* rtheta_i = (double *)R_alloc(1, sizeof(double)); */
	rtheta_i =  REAL(VECTOR_ELT(theta, i)); 
	/* rtheta_i = &rtheta[i]; */

    loglik[i] = logsum(P, ra_N, nP, na_N, rbase_i, rtheta_i);
  }
    
  *rans = -dvecsum(loglik, nY[0]);

  UNPROTECT(1);

  return ans;
}

/* chindex function updates the index by omiting "omit"*/
void chindex(int* seq, int nseq, int a, int omit)
{
  int i, j;

  j = 0;
  for (i = 0; i < nseq; i++) {
    if (i == omit) {
	  j++;
	} 
    seq[i] = j + a;
    j++;
  }
}

/* makeimat tries to find a sub-matrix of the matrix old  */
void makeimat(int *old, int *new, int *no, int *nn,
	      int radd, int cadd, int romit, int comit)
{
  int i, j, r, c;
  for (i = 0, r = 0; i < nn[0]; i++, r++) {
    if (i == romit)
      r++;
    for (j = 0, c = 0; j < nn[1]; j++, c++) {
      if (j == comit)
	    c++;
      new[i + j*nn[0]] = old[r + radd + (c + cadd)*no[0]];
    }
  }
}

/* rowreduce does the operation of subtracting the last
element from each element in a vector x */
void rowreduce(double *x, int nx)
{
  int i;
  for (i = 0; i < nx; i++)
    x[i] -= x[nx-1];
}


/* gradient function calculates the gradient of the log-likelihood 
function with respect to the probability parameters */
SEXP gradient(SEXP x, SEXP AllPtrue, SEXP divsel, SEXP Y,
	       SEXP a, SEXP aa, SEXP MAX, SEXP base, SEXP theta)
{
  int nx, ndivsel, nAll[2], nY[2], nP[2], nP1[2], na_N[2], ngrad;
  int *rY, *rdivsel, *rMAX, *ra_N, *rbase_i;
  int i, j, k, l, cnt, N;
  int *rowAll, *rowP, *colP;
  /* int *rowra_N, *colra_N; */
  /* double *rtheta; */

  double *rx, *rAllPtrue, *rans, *rtheta_i;
  double *AllP, *P, *P1, *loglik, *grad;
 
  SEXP ans;

  nx = length(x);
  ndivsel = length(divsel);
  nAll[0] = DIMS(AllPtrue)[0];
  nAll[1] = DIMS(AllPtrue)[1];
  nY[0] = DIMS(Y)[0];
  nY[1] = DIMS(Y)[1];

  rx = REAL(x);
  rAllPtrue = REAL(AllPtrue);
  /* rtheta = REAL(theta); */
  rdivsel = INTEGER(divsel);
  rY = INTEGER(Y);
  rMAX = INTEGER(MAX);

  loglik = (double *)R_alloc(nY[0], sizeof(double));
  /* div_sel = INTEGER(VECTOR_ELT(division, rsel[0]-1))[0]; */

  AllP = (double *)R_alloc(nAll[0]*nAll[1], sizeof(double));
  for (i = 0; i < nAll[0]; i++)
    for (j = 0; j < nAll[1]; j++)
      AllP[i + j*nAll[0]] = rAllPtrue[i + j*nAll[0]];
 

  for(j = 0; j < ndivsel ;j++) {
      for (i = 0; i < nAll[0] - 1; i++) {
		  AllP[i + (rdivsel[j] - 1)*nAll[0]] = rx[i];
	  }
      AllP[i + (rdivsel[j] - 1)*nAll[0]] = 1 - dvecsum(rx, nx);
  }

  ngrad = nY[1];
  grad = (double *)R_alloc(ngrad, sizeof(double));
  for (i = 0; i < ngrad; i++)
    grad[i] = 0.0;

  PROTECT(ans = allocVector(REALSXP, ngrad - 1));
  rans = REAL(ans);

  for (i = 0; i < nY[0]; i++) {

    /* get the number of mutations in sample i */
    N = irowsum(rY, nY, i);

    nP[0] = N;
    nP[1] = N;
    P = (double *)R_alloc(nP[0]*nP[1], sizeof(double));
    rowAll = (int *)R_alloc(nP[0], sizeof(int));
	
    /* get the index of mutated genes in sample i */	
    l = 0;
    for (j = 0; j < nY[1]; j++){
      for (k = 0; k < rY[i + j*nY[0]]; k++) {
		  rowAll[l++] = j;
	  }
	}
	/* consider sub-matrix of full probability matrix */
    dsubmat(AllP, P, nAll, nP, rowAll, NULL);

    na_N[0] = DIMS(VECTOR_ELT(a, N - 1))[0]; 
    na_N[1] = DIMS(VECTOR_ELT(a, N - 1))[1];
    ra_N = INTEGER(VECTOR_ELT(a, N - 1));
	rbase_i = INTEGER(VECTOR_ELT(base, i));
	/* rtheta_i = (double *)R_alloc(1, sizeof(double)); */
	rtheta_i =  REAL(VECTOR_ELT(theta, i)); 
	/* rtheta_i = &rtheta[i]; */
	/*rtheta_i = veci(rtheta, i); */

    if (rdivsel[0] <= N) {

      if (N > 1) {

	    loglik[i] = logsum(P, ra_N, nP, na_N, rbase_i, rtheta_i);
		
	    nP1[0] = nP[0] - 1;
	    nP1[1] = nP[1] - 1;
		/* R_alloc function ? */
	    P1 = (double *)R_alloc(nP1[0]*nP1[1], sizeof(double));
	    rowP = (int *)R_alloc(nP1[0], sizeof(int));
	    colP = (int *)R_alloc(nP1[1], sizeof(int));

	    cnt = 0;
	    for (j = 0; j < nP[0]; j++) {
	      cnt++;

	      for(l = rdivsel[0]; l <= imin2(N, rdivsel[ndivsel - 1]); l++) {
			/* compute gradient by looking at structure of log-likelihood function */ 
            /* change index by omitting one row "cnt" and one column "l" */			
	        chindex(rowP, nP1[0], 0, cnt - 1);
	        chindex(colP, nP1[1], 0, l - 1);
	        dsubmat(P, P1, nP, nP1, rowP, colP);

	        if (l > rMAX[0]) {
	          /* na_N1[0] = DIMS(VECTOR_ELT(a, N - 2))[0];
	          na_N1[1] = DIMS(VECTOR_ELT(a, N - 2))[1];
		      ra_N1 = INTEGER(VECTOR_ELT(a, N - 2));
		      grad[rowAll[j]] +=
		        exp(logsum(P1, ra_N1, nP1, na_N1) - loglik[i])/(N - rMAX[0]); */
			  grad[rowAll[j]] +=
		        exp(logsumomit(P, ra_N, nP, na_N, rbase_i, rtheta_i, cnt - 1, l - 1) - loglik[i])/(N - rMAX[0]);
				
	        } else {
		      /* na_N1[0] = DIMS(VECTOR_ELT(aa, N - 2))[0];
		      na_N1[1] = DIMS(VECTOR_ELT(aa, N - 2))[1];
		      ra_N1 = INTEGER(VECTOR_ELT(aa, N - 2));
		      grad[rowAll[j]] +=
		        exp(logsum(P1, ra_N1, nP1, na_N1) - loglik[i]); */
		      grad[rowAll[j]] +=
		        exp(logsumomit(P, ra_N, nP, na_N, rbase_i, rtheta_i, cnt - 1, l - 1) - loglik[i]);
				
			}
		  }
		}
	    rowreduce(grad, ngrad);
		
	  } else {
	    for (j = 0; j < nY[1]; j++)
	      if (rY[i + j*nY[0]] == 1)
	        grad[j] += 1./P[0];

	    rowreduce(grad, ngrad);
	  }
	}
  }
    
  for (i = 0; i < ngrad - 1; i++)
    rans[i] = -grad[i];
    
  UNPROTECT(1);
  return ans;
}



